import os
from abc import abstractmethod

import tensorflow as tf
import numpy as np

from .agent import Agent


class RLAgent(Agent):
    """Abstract base class for Agents that play and learn on pyloa.environment.PagingEnvironment instances.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        learning_rate (float): learning_rate of agent's model
        epsilon (float): agent's current chance to exploit his knowledge, else explore action_space
        epsilon_delta (float): increment value of epsilon for each increment call
        epsilon_max (float): maximum value of epsilon
        save_file (str): location at which agent's checkpoints are stored at

    Attributes:
        _learning_rate (float): learning_rate of agent or agent's model
        _epsilon (float): agent's current chance to exploit his knowledge, else explore action_space
        _epsilon_delta (float): increment value of epsilon for each increment call
        _epsilon_max (float): maximum value of epsilon
        _save_file (str): location at which agent's checkpoints are stored at
        _sess (tf.python.client.session.Session): tensorflow session handle
        _global_step (tf.python.ops.variables.RefVariable): global train steps for agent's model
    """
    def __init__(self,
                 state_dim,
                 action_dim,
                 learning_rate,
                 epsilon,
                 epsilon_delta,
                 epsilon_max=1.0,
                 save_file=None,
                 **kwargs):
        super().__init__(state_dim=state_dim, action_dim=action_dim, **kwargs)
        # params
        self._learning_rate = learning_rate
        self._epsilon = epsilon
        self._epsilon_delta = epsilon_delta
        self._epsilon_max = epsilon_max
        self._save_file = save_file
        # tensorflow
        self._sess = tf.Session()
        self._global_step = tf.Variable(0, name='global_step', trainable=False)

    @abstractmethod
    def _build_model(self):
        """Build the model: e.g. a tensorflow computational graph.

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def act(self, state):
        """Agent decides on an action based on information given with state.

        Args:
            state (np.ndarray with shape=(_state_dim,)): input state

        Returns:
            int: action to be taken

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def train(self):
        """Agent trains model.

        Raises:
            NotImplementedError: if abstract method is to be invoked

        Returns:
            float: loss of train step
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def store(self, state, action, reward, next_state):
        """Store quadruple of experience.

        Args:
            state (np.ndarray with shape=(_state_dim,): observed initial state before action
            action (int): action selected for state
            reward (float): reward from environment for taking action in state
            next_state (np.ndarray with shape=(_state_dim,)): observed state after action
        """
        raise NotImplementedError("Must be implemented in subclass.")

    def save(self, file_path=None):
        """Save agent's model with _global_step as suffix on persistent storage.

        Args:
            file_path (str): existing path to file
        """
        file_path = file_path if file_path is not None else self._save_file
        saver = tf.train.Saver()
        path = saver.save(self._sess, file_path, global_step=self._global_step)
        self._logger.info("Saved model {} in {}".format(self.__class__.__name__, path))

    def load(self, file_path=None):
        """Load agent's last checkpoint from persistent storage.

        Args:
            file_path (str): existing path to checkpoint
        """
        file_path = file_path if file_path is not None else self._save_file
        _cp_file = tf.train.latest_checkpoint(os.path.dirname(file_path))
        saver = tf.train.Saver()
        try:
            saver.restore(self._sess, _cp_file)
            self._logger.info("Loaded model {} from {}".format(self, _cp_file))
        except Exception as e:
            self._logger.error("Couldn't load model {} from file='{}': {}".format(self, _cp_file, str(e)))
            self._logger.info("model {} starts with default initialization".format(self))

    def increment_epsilon(self):
        """Increase agent's epsilon by _epsilon_delta without exceeding _epsilon_max."""
        self.epsilon = min(self._epsilon + self._epsilon_delta, self._epsilon_max)

    def _get_epsilon(self):
        """float: current chance of agent to exploit his current knowledge, else explore action_space"""
        return self._epsilon

    def _set_epsilon(self, value):
        assert isinstance(value, (float, np.float)), "epsilon must be of type '{}' or '{}' but found type='{}' instead"\
            .format(float, np.float, type(value))
        assert 0.0 <= value <= 1.0, "epsilon must be in interval [0.0, 1.0] but found value='{}' instead".format(value)
        self._epsilon = value

    epsilon = property(fset=_set_epsilon, fget=_get_epsilon, fdel=None)



import numpy as np

from .environment import PagingEnvironment


# Todo @schoenitz documentation
class PermutatedPagingEnvironment(PagingEnvironment):
    def __init__(self,
                 sequence_size,
                 cache_size,
                 num_pages,
                 include_request=False,
                 include_pagefault=False,
                 include_cache=False,
                 include_time_in_cache=False,
                 include_time_since_touched=False,
                 include_histogram=False,
                 dtype=np.int32,
                 **kwargs):
        super().__init__(
            sequence_size=sequence_size,
            cache_size=cache_size,
            num_pages=num_pages,
            include_request=include_request,
            include_pagefault=include_pagefault,
            include_cache=include_cache,
            include_time_in_cache=include_time_in_cache,
            include_time_since_touched=include_time_since_touched,
            include_histogram=include_histogram,
            dtype=dtype,
            **kwargs
        )

        _cache_perm = {i + 1: i + 1 for i in range(self._cache_size)}
        _val_perm = {i: i for i in range(self._num_pages + 1)}
        self._permutation = {**_cache_perm, **_val_perm}

    def _update_permutation(self, pos, val):
        """We apply a transposition on the permutation itself"""
        if pos not in self._permutation.keys():
            raise ValueError(pos, "(as position) is not a valid key in the permutation:", self._permutation)
        if val not in self._permutation.keys():
            raise ValueError(val, "(as value) is not a valid key in the permutation:", self._permutation)
        self._permutation[pos], self._permutation[val] = self._permutation[val], self._permutation[pos]

    def _get_val_after_permutation(self, val):
        if val not in self._permutation.keys():
            raise ValueError(val, "is not a valid key in the permutation:", self._permutation)
        return self._permutation[val]

    def _get_histogram(self):
        return self._histogram

    def _get_time_in_cache(self):
        return self._time_in_cache

    def _get_time_since_touched(self):
        return self._time_since_touched

    def _calculate_cache_state_complexity(self):
        if self._include_cache:
            return 2 ** self._cache_size
        else:
            return 1

    def _calculate_time_in_cache_state_complexity(self):
        pass

    def _calculate_time_since_touched_state_complexity(self):
        pass

    def _calculate_histogram_state_complexity(self):
        pass

    def _is_pagefault(self):
        return self._get_val_after_permutation(self.request) not in self.cache

    def step(self, action, vectorized_state=True):
        # check bounds (0 means 'no action',  interval [1, cache_size] represents indices (offset by one))
        if not 0 <= action <= self._cache_size:
            raise ValueError("Out ouf bounds")
        used_indices = self.used_indices

        permutated_request = self._get_val_after_permutation(self.request)

        # no action (no flush, current_page expected by agent to exist in cache)
        if action == 0:
            # agent is right, page exists in cache -> update structures
            if not self.is_pagefault:
                idx = permutated_request - 1
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._histogram[idx] += 1
                cost = 0  # no costs, since agent correctly identified that current_page is in cache
            # agent is incorrect, page doesn't exist in cache
            else:
                self._time_since_touched[used_indices] += 1
                self._time_in_cache[used_indices] += 1
                cost = 1000  # high costs, since agent falsely expected current_page to be in cache
        # agent choose action to flush index to place current_page
        else:
            # agent tries to flush an index, even though current_page is already in cache
            if not self.is_pagefault:
                idx = np.where(self._cache == permutated_request)[0][0]
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._histogram[idx] += 1
                cost = 500
            # agent is right, page doesn't exist in cache -> flush selected action index
            else:
                idx = action - 1
                # agent decides to place current_page in empty cache slot -> no flush
                if 0 in self._cache and self._cache[idx] == 0:
                    cost = 0
                # agent decide to place current_page in a non-empty index, even though there is free space
                elif 0 in self._cache and self._cache[idx] != 0:
                    cost = 250
                # normal flush
                else:
                    cost = 10
                # self._cache[idx] = self.request
                self._cache[idx] = idx + 1
                self._update_permutation(idx + 1, permutated_request)
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._time_in_cache[idx] = 1
                self._histogram[idx] = 1
        # inc step
        self._num_steps += 1
        return self.get_state(vectorized_state=vectorized_state), cost

    histogram = property(fset=None, fget=_get_histogram, fdel=None)
    time_in_cache = property(fset=None, fget=_get_time_in_cache, fdel=None)
    time_since_touched = property(fset=None, fget=_get_time_since_touched, fdel=None)
    is_pagefault = property(fset=None, fget=_is_pagefault, fdel=None)



import logging
from abc import abstractmethod, ABCMeta


class Agent(metaclass=ABCMeta):
    """Abstract base class for Agents.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        name (str): depicts a name for the agent

    Attributes:
        _logger (logging.Logger): logger instance for each subclassed instance
        _state_dim (int): defines input shape=(_state_dim,) of input states
        _action_dim (int): defines the number of different actions the agent can take
        _name (str): depicts a name for the agent
    """
    def __init__(self, state_dim, action_dim, name=None):
        self._name = self.__class__.__name__ if name is None else name
        self._state_dim = state_dim
        self._action_dim = action_dim
        self._logger = logging.getLogger(__name__)
        self._logger.info("initializing {}".format(self))

    @abstractmethod
    def act(self, state):
        """Agent decides on an action based on information given with state.

        Args:
            state (np.ndarray with shape=(_state_dim,) or dict): input state

        Returns:
            int: action to be taken

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def play_instance(self, environment, episode, is_train, writer, invert_cost=True):
        """Agent plays an instance on environment untill environment terminates.

        Args:
            environment (pyloa.environment.Environment): environment to play on
            episode (int): number depicting how many episodes agent played so far
            is_train (bool): depicts if agent should train (True) or just evaluate (False)
            writer (tf.summary.FileWriter): tensorboard logging for playing a instance
            invert_cost (bool): depicts if costs are inverted after calculation (cost = negative reward)

        Returns:
            int: action to be taken

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    def __str__(self):
        """str: string representation of agent"""
        return self._name



# define abstract alias imports for other modules
from .environment import Environment
from .environment import PagingEnvironment
# define alias imports for other modules
from .permutatedenv import PermutatedPagingEnvironment
from .normalizedenv import NormalizedPagingEnvironment
from .defaultenv import DefaultPagingEnvironment
from .vertexenv import VertexEnvironment



from setuptools import setup, find_packages
import os

# load info from files
setup_directory = os.path.abspath(os.path.dirname(__file__))
readme_file = os.path.join(setup_directory, 'README.md')
version_file = os.path.join(setup_directory, 'pyloa', '_version.py')
requirements_file = os.path.join(setup_directory, 'requirements.txt')
__version__ = None

# load long description for PyPi
with open(readme_file, encoding='utf-8', mode='r') as f:
    __long_description__ = f.read()

# get version number
with open(version_file, encoding='utf-8', mode='r') as fd:
    exec(fd.read())

# get dependencies
with open(requirements_file, encoding='utf-8', mode='r') as f:
    __install_requires__ = [i for i in f.read().strip().split('\n')]

# PyPi classifiers
__classifiers__ = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Education",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
]

setup(
    name="pyloa",
    version=__version__,
    license="MIT",
    author="PyLoa Developers",
    author_email="pyloadevelopers@gmail.com",
    maintainer="PyLoa Developers",
    maintainer_email="pyloadevelopers@gmail.com",
    packages=['pyloa'] + ['pyloa.' + i for i in find_packages('pyloa')],
    provides=['pyloa'],
    python_requires='>=3.5',
    classifiers=__classifiers__,
    install_requires=__install_requires__,
    description="PyLoa: Learning on-line Algorithms with Python",
    long_description=__long_description__,
    long_description_content_type='text/markdown',
    url="https://github.com/pyloa/PyLoa",
    project_urls={
        "Bug Tracker": "https://github.com/pyloa/PyLoa/issues",
        "Documentation": "https://github.com/pyloa/PyLoa/tree/master/pyloa",
        "Source Code": "https://github.com/pyloa/PyLoa/tree/master/pyloa",
    },
    platforms=['any'],
    entry_points={'console_scripts': ['pyloa=pyloa._main:main'], },
    keywords='on-line algorithms, paging, vertex coloring, machine learning',
    zip_safe=False
)



from abc import abstractmethod

import tensorflow as tf
import numpy as np

from .rlagent import RLAgent
from pyloa.utils.memory import Memory, ReplayMemory, PrioritizedReplayMemory


class QAgent(RLAgent):
    """Abstract base class for Agents that play and learn on pyloa.environment.PagingEnvironment with Q-Learning on a
    replay memory according to Mnih et al. in https://storage.googleapis.com/deepmind-media/dqn/DQNNaturePaper.pdf

    QAgent defines an abstract class for Deep-Q-Networks (DQN); implementations of QAgent have to implement the function
    _build_q_network, which has to return the output of the network (Q_Values). QAgent invokes _build_q_network twice
    while building the model; firstly, to create the Q-target Network and secondly to create the Q-eval Network - both
    have the same structure according to the child's implementation. Q-target's variables are frozen and won't be
    updated during _train_op; only Q-eval will be trained. Every _replace_threshold-many train steps the operation
    _replace_params is invoked, which replaces Q-target's variables with Q-eval's variables. Each entry within the
    agent's replay pyloa.utils.Memory is a quadruple of (state, action, reward, next_state); each train step draws
    uniformly a random a sample of size _sample_size from replay memory. Each sample is batched with batch_size. The
    _loss function for training is defined as:

        _loss = (reward + _discount_factor * max(q_target(next_state)) - q_eval(state)) ** 2

    whereas q_eval's entry at index=action (from memory) is used and q_target's entry at index=argmax is used. The _loss
    function is L2-regularized with _l2_scale for all variables in Q-eval with:

        for var in q_eval_variables:
             _loss += _l2_scale * tf.nn.l2_loss(var)

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        learning_rate (float): learning_rate of agent's model
        epsilon (float): agent's current chance to exploit his knowledge, else explore action_space
        epsilon_delta (float): increment value of epsilon for each increment call
        epsilon_max (float): maximum value of epsilon
        discount_factor (float): value at which future Q-target Values are discounted
        memory_size (int): total size of agent's replay memory
        sample_size (int): size of sample randomly drawn from memory to train on per train step
        batch_size (int): size of each batch a sample for training is batched with
        l2_scale (float): L2-Regularization scale for eval network
        replace_threshold (int): depicts after how many train steps Q-target's variables are replaced with Q-eval's
        save_file (str): location at which agent's checkpoints are stored at

    Attributes:
        _discount_factor (float): value at which Q-target Values are discounted
        _memory_size (int): total size of agent's replay memory
        _sample_size (int): size of sample randomly drawn from memory to train on per train step
        _batch_size (int): size of each batch a sample for training is batched with
        _l2_scale (float): L2-Regularization scale for eval network
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _replace_step (int): counting variable from 0 to _replace_threshold
        _memory (pyloa.utils.Memory): replay memory to store in and draw from
        _state (tf.Tensor): placeholder with shape=(?, _state_dim) for _batch_size many states
        _next_state (tf.Tensor): placeholder shape=(?, _state_dim) for _batch_size many next_states
        _reward (tf.Tensor): placeholder shape=(?,) for _batch_size many rewards
        _action (tf.Tensor): placeholder shape=(?,) for _batch_size many actions
        _loss (tf.Tensor): as defined above in __class__.__doc__
        _train_op (tf.Operation): minimizes _loss with RMSPropOptimizer with respect to _learning_rate
        _eval_outs: (tf.Tensor): Q-eval Values Tensor for inference on model (invoked from act)
        _eval_params (list of tf.Variable): Q-eval Network variables
    """
    def __init__(self,
                 state_dim,
                 action_dim,
                 learning_rate,
                 epsilon,
                 epsilon_delta,
                 epsilon_max,
                 discount_factor,
                 memory_size,
                 sample_size,
                 batch_size,
                 l2_scale,
                 replace_threshold,
                 seed=None,
                 save_file=None,
                 **kwargs):
        super().__init__(
            state_dim=state_dim,
            action_dim=action_dim,
            learning_rate=learning_rate,
            epsilon=epsilon,
            epsilon_delta=epsilon_delta,
            epsilon_max=epsilon_max,
            save_file=save_file,
            **kwargs
        )
        # params
        self._discount_factor = discount_factor
        self._memory_size = memory_size
        self._sample_size = sample_size
        self._batch_size = batch_size
        self._l2_scale = l2_scale
        self._replace_threshold = replace_threshold
        # vars
        self._rnd_state = np.random.RandomState(seed=seed)
        self._replace_step = 0
        if 'memory' in kwargs and 'type' in kwargs['memory']:
            if not issubclass(kwargs['memory']['type'], Memory):
                raise RuntimeError("optional memory implementation must subclass '{}'".format(Memory))
            self._memory = kwargs['memory']['type'](memory_size=self._memory_size, **kwargs['memory'])
        else:
            self._memory = ReplayMemory(memory_size=self._memory_size)
        self._eval_params = None
        # tensorflow tensors
        if isinstance(self._memory, PrioritizedReplayMemory):
            self._isw = tf.placeholder(dtype=tf.float32, shape=[None, ], name='importance_sampling_weights')
        self._state = tf.placeholder(dtype=tf.float32, shape=[None, self._state_dim], name='state')
        self._next_state = tf.placeholder(dtype=tf.float32, shape=[None, self._state_dim], name='next_state')
        self._reward = tf.placeholder(dtype=tf.float32, shape=[None, ], name='reward')
        self._action = tf.placeholder(dtype=tf.int32, shape=[None, ], name='action')
        self._loss = None
        self._eval_outs = None
        # tensorflow ops
        self._train_op = None

    def _build_model(self, **kwargs):
        """Builds the Deep Q-Network.

        Construct two equally structured Q-Networks (q_eval and q_target) according to child class' implementation.
        Defines _loss function, applies L2-Regulrization to _loss, defines optimizer and train operation"""
        with tf.variable_scope("model", reuse=tf.AUTO_REUSE):
            with tf.variable_scope("q_target", reuse=tf.AUTO_REUSE):
                target_outs = self._build_q_network(input_tensor=self._next_state, reuse=tf.AUTO_REUSE, **kwargs)
                target_params = tf.get_collection(key=tf.GraphKeys.GLOBAL_VARIABLES, scope=tf.get_variable_scope().name)
                target_outs = tf.stop_gradient(target_outs)
            with tf.variable_scope("q_eval", reuse=tf.AUTO_REUSE):
                eval_outs = self._build_q_network(input_tensor=self._state, reuse=tf.AUTO_REUSE, **kwargs)
                eval_params = tf.get_collection(key=tf.GraphKeys.GLOBAL_VARIABLES, scope=tf.get_variable_scope().name)
            with tf.variable_scope('replace', reuse=tf.AUTO_REUSE):
                self._replace_params = [tf.assign(t, e) for t, e in zip(target_params, eval_params)]
            with tf.variable_scope("merge", reuse=tf.AUTO_REUSE):
                _batch_size = tf.shape(self._state)[0]
                _range = tf.range(_batch_size)
                _a_indices = tf.stack([_range, self._action], axis=1)
                q_eval_actions = tf.gather_nd(params=eval_outs, indices=_a_indices)
                q_target_actions = self._reward + self._discount_factor * tf.reduce_max(target_outs, axis=1)
                self._td_error = tf.abs(tf.subtract(q_target_actions, q_eval_actions))
            with tf.variable_scope("loss", reuse=tf.AUTO_REUSE):
                if isinstance(self._memory, PrioritizedReplayMemory):
                    self._loss = self._isw * tf.square(self._td_error)
                else:
                    self._loss = tf.square(self._td_error)
                for var in eval_params:
                    self._loss = self._loss + self._l2_scale * tf.nn.l2_loss(var)
            with tf.variable_scope("opt", reuse=tf.AUTO_REUSE):
                optim = tf.train.RMSPropOptimizer(learning_rate=self._learning_rate)
                self._train_op = optim.minimize(self._loss, global_step=self._global_step, var_list=eval_params)
        # variable hooks
        self._eval_outs = eval_outs
        self._eval_params = eval_params

    @abstractmethod
    def _build_q_network(self, input_tensor, reuse=tf.AUTO_REUSE, **kwargs):
        """Builds a Q-Network for Q-Value estimation.

        Args:
            input_tensor (tf.Tensor): accordingly _state or _next_state tensor
            reuse (enum _ReuseMode): tensorflow's reusemode

        Returns:
            tf.Tensor: q_value output of Q-Network

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    def get_params(self):
        """dict: returns the currently used variables in Q-eval Network"""
        var_dict = {}
        for var in self._eval_params:
            var_dict[var.name] = self._sess.run(var)
        return var_dict

    def act(self, state, **kwargs):
        """QAgent selects an action by either exploiting his knowledge or by exploring his action space. Exploitation
        depicts the inference on Q-eval with state, whereas Exploration depicts a uniformly drawn random action.

        Args:
            state (np.ndarray with shape=(_state_dim,)): input state

        Returns:
            int: action to be taken
        """
        if self._rnd_state.uniform(low=0.0, high=1.0) < self._epsilon:      # epsilon-greedy exploitation
            q_action = self._sess.run(self._eval_outs, feed_dict={self._state: [state]})
            action = np.argmax(q_action[0])
        else:                                                               # (1-epsilon) exploration
            action = self._rnd_state.randint(0, self._action_dim)
        return action

    def train(self, train_writer=None, **kwargs):
        """QAgent trains on a sample of replay memory.

        Firstly, requests a sample of size _sample_size from _memory. Secondly, batches through sample with _batch_size.
        Loss values for each batch are appended and returned.

        Args:
            train_writer (tf.summary.FileWriter): writer instance for tensorboard

        Returns:
            list of float: loss of training on drawn samples
        """
        # check for replace step
        if self._replace_threshold == self._replace_step:
            self._sess.run(self._replace_params)
            self._logger.info("replaced q_target's variables with q_eval's variables")
            self._replace_step = 0

        losses = []
        feed_dict = dict()
        # batch through sample
        for step in range(int(self._sample_size / self._batch_size)):

            # get train samples
            if isinstance(self._memory, PrioritizedReplayMemory):
                tree_indices, feed_dict[self._isw], experiences = self._memory.sample(sample_size=self._batch_size)
            else:
                experiences = self._memory.sample(sample_size=self._batch_size)

            # build feed dict
            feed_dict[self._state], feed_dict[self._action], feed_dict[self._reward], feed_dict[self._next_state] =\
                zip(*experiences)

            # train step
            _, _td_error, _loss = self._sess.run([self._train_op, self._td_error, self._loss], feed_dict=feed_dict)

            # save losses
            losses.append(_loss)

            # update priorities
            if isinstance(self._memory, PrioritizedReplayMemory):
                self._memory.update(tree_indices=tree_indices, td_errors=_td_error)
                self._memory.increment_beta()

        # replace step inc
        self._replace_step += 1
        return losses

    def store(self, state, action, reward, next_state):
        """Store quadruple of experience in replay memory.

        Args:
            state (np.ndarray with shape=(_state_dim,): observed initial state before action
            action (int): action selected for state
            reward (float): reward from environment for taking action in state
            next_state (np.ndarray with shape=(_state_dim,)): observed state after action
        """
        self._memory.store(experience=(state, action, reward, next_state))



import os
import pickle as pk

import numpy as np

from .rlagent import RLAgent
from .pagingagent import PagingAgent


class QTableAgent(RLAgent, PagingAgent):
    """A simple Q-Learning table agent with dict as a table container. This is a naive implementation for demonstration
    purposes only. Use QTableAgent only on toy examples, elsewise you will run out of memory.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        learning_rate (float): learning_rate of agent's model
        discount_factor (float): value at which future Q-target Values are discounted
        epsilon (float): agent's current chance to exploit his knowledge, else explore action_space
        epsilon_delta (float): increment value of epsilon for each increment call
        epsilon_max (float): maximum value of epsilon
        save_file (str): location at which agent's checkpoints are stored at

    Attributes:
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _discount_factor (float): value at which Q-target Values are discounted
        _q_table (dict): key type=str as state, value type=np.ndarray shape=(_action_dim,) as Q-Values for each action
    """
    # noinspection PyUnusedLocal
    def __init__(self,
                 state_dim,
                 action_dim,
                 learning_rate,
                 discount_factor,
                 epsilon=0.90,
                 epsilon_delta=50e-4,
                 epsilon_max=1.0,
                 save_file=None,
                 seed=None,
                 **kwargs):
        super().__init__(
            state_dim=state_dim,
            action_dim=action_dim,
            learning_rate=learning_rate,
            epsilon=epsilon,
            epsilon_delta=epsilon_delta,
            epsilon_max=epsilon_max,
            save_file=save_file
        )
        # params
        self._rnd_state = np.random.RandomState(seed=seed)
        self._discount_factor = discount_factor
        self._q_table = None
        # init
        self._build_model()

    def _build_model(self):
        """initialize a dictionary used as a lookup table as model."""
        self._q_table = dict()

    def act(self, state):
        """QTableAgent decides on an action based on information given with state.

        Args:
            state (np.ndarray, shape=(_state_dim,), dtype=np.int32): input state

        Returns:
            int: action to be taken
        """
        _h_state = str(state)                                                           # np.ndarray -> str
        if _h_state not in self._q_table:                                               # state not seen yet
            self._q_table[_h_state] = np.zeros(self._action_dim, dtype=np.int32)        # initial Q-Values are 0
            action = self._rnd_state.randint(0, self._action_dim)                       # explore random action
        elif self._rnd_state.uniform(low=0.0, high=1.0) < self._epsilon:                # e-greedy exploitation
            action = np.argmax(self._q_table[_h_state])                                 # pick best action
        else:                                                                           # (1-e) exploration
            action = self._rnd_state.randint(0, self._action_dim)                       # explore random action
        return action

    # noinspection PyMethodOverriding
    def train(self, state, action, reward, next_state):
        """QTableAgent trains.

        Args:
            state (np.ndarray, shape=(_state_dim,), dtype=np.int32): observed initial state before action
            action (int): action selected for state
            reward (float): reward from environment for taking action in state
            next_state: (np.ndarray, shape=(_state_dim,), dtype=np.int32): observed state after action

        Returns:
            int: number of states explored by QTableAgent
        """
        _h_state, _h_nstate = str(state), str(next_state)                               # np.ndarray -> str
        if _h_nstate not in self._q_table:                                              # next state not seen yet
            self._q_table[_h_nstate] = np.zeros(self._action_dim, dtype=np.int32)       # initial Q-Values are 0
        # update Q-Values for action
        self._q_table[_h_state][action] = self._q_table[_h_state][action] + self._learning_rate * \
            (reward + self._discount_factor * np.max(self._q_table[_h_nstate]) - self._q_table[_h_state][action])
        return len(self._q_table)

    def save(self, file_path=None):
        """Save agent's Q-Table on persistent storage.

        Args:
            file_path (str): existing path to file
        """
        file_path = file_path if file_path is not None else self._save_file
        with open(file_path + '.pkl', 'wb') as f:
            pk.dump(self._q_table, f, pk.HIGHEST_PROTOCOL)
        self._logger.info("Saved model {} in {}".format(self, file_path))

    def store(self, state, action, reward, next_state):
        """QTableAgent does not have a memory to store in."""
        pass

    def load(self, file_path=None):
        """Load agent's Q-Table from persistent storage.

        Args:
            file_path (str): existing path to picked Q-Table
        """
        file_path = file_path if file_path is not None else self._save_file
        directory = os.path.dirname(file_path) if not os.path.isdir(file_path) else file_path
        try:
            with open(os.path.join(directory, self.__class__.__name__ + '.pkl'), 'rb') as f:
                self._q_table = pk.load(f)
            self._logger.info("Loaded model {} from {}".format(self, file_path))
        except Exception as e:
            self._logger.error("Couldn't load model {} from file='{}': {}".format(self, file_path, str(e)))
            self._logger.info("model {} starts with default initialization".format(self))
